"""Utility functions for LSST JupyterLab notebook environment
"""
import bokeh
import os
from urllib.parse import urljoin


def format_bytes(n):
    """ Format bytes as text

    >>> format_bytes(1)
    '1 B'
    >>> format_bytes(1234)
    '1.23 kB'
    >>> format_bytes(12345678)
    '12.35 MB'
    >>> format_bytes(1234567890)
    '1.23 GB'
    >>> format_bytes(1234567890000)
    '1.23 TB'
    >>> format_bytes(1234567890000000)
    '1.23 PB'

    (taken from dask.distributed, where it is not exported)
    """
    if n > 1e15:
        return '%0.2f PB' % (n / 1e15)
    if n > 1e12:
        return '%0.2f TB' % (n / 1e12)
    if n > 1e9:
        return '%0.2f GB' % (n / 1e9)
    if n > 1e6:
        return '%0.2f MB' % (n / 1e6)
    if n > 1e3:
        return '%0.2f kB' % (n / 1000)
    return '%d B' % n


def get_proxy_url(port):
    """Assumes the LSST environment and that nbserverproxy is running.
    Returns the external address that will resolve to the supplied port
    """
    if not port:
        return None

    base_url = os.environ.get('EXTERNAL_URL') or 'http://localhost:8888'
    service_url_path = os.environ.get('JUPYTERHUB_SERVICE_PREFIX')
    proxy_url_path = 'proxy/{port}'.format(port=port)
    return urljoin(urljoin(base_url, service_url_path), proxy_url_path)


def get_hostname():
    """Utility function to return hostname or, failing that, "localhost".
    """
    return os.environ.get('HOSTNAME') or 'localhost'


def show_with_bokeh_server(obj):
    """Method to wrap bokeh with proxy URL
    """
    def jupyter_proxy_url(port):
        """Construct proxy URL from environment
        """
        # If port is None we're asking about the URL
        # for the origin header.
        return get_proxy_url(port) or '*'

    bokeh.io.show(obj, notebook_url=jupyter_proxy_url)



"""
Collection of notebook utilities
"""
from .clusterproxy import ClusterProxy
from .forwarder import Forwarder
from .lsstdaskclient import LSSTDaskClient
from .utils import format_bytes, get_proxy_url, get_hostname, \
    show_with_bokeh_server

__all__ = [ClusterProxy, Forwarder, LSSTDaskClient, format_bytes,
           get_proxy_url, get_hostname, show_with_bokeh_server]



import ipaddress
import logging
from .forwarder import Forwarder
from .utils import get_proxy_url, get_hostname
from dask.distributed import Client


def _remove_prefix(s, prefix):
    return s[len(prefix):] if s.startswith(prefix) else s


class ClusterProxy(object):
    """Provides a proxy service to map a local port to a worker node's
    dashboard, which should be on its "bokeh" service.

    This allows us to proxy to the worker even though the k8s network is not
    accessible externally.

    It must be created with an instance of a dask.distributed.Client
    as its argument.

    This would work better if k8s workers had a bokeh service, and if the
    bokeh service told you how to get to its dashboard page once you have
    the host and port.
    """
    client = None
    cluster = None
    ioloop = None
    workers = {}

    def __init__(self, client):
        if not isinstance(client, Client):
            estr = "'client' argument must be dask.distributed.Client!"
            raise RuntimeError(estr)
        self.client = client
        self.logger = logging.getLogger(__name__)
        self.cluster = client.cluster
        self.ioloop = client.io_loop
        self.scheduler_url = None
        port = client.cluster.scheduler.identity()["services"].get('bokeh')
        if port:
            self.scheduler_url = get_proxy_url(port) + "/status"
        self.refresh_workers()

    def refresh_workers(self):
        """Rebuild current worker map from actual state.
        """
        current_workers = self.cluster.scheduler.identity().get('workers')
        current_workerlist = [_remove_prefix(x, "tcp://")
                              for x in current_workers]
        removed_workers = []
        for worker_id in self.workers:
            if worker_id not in current_workerlist:
                # Get rid of removed workers
                self._remove_worker_proxy(worker_id)
                removed_workers.append(worker_id)
        for worker_id in removed_workers:
            del self.workers[worker_id]
        for worker_id in current_workerlist:
            if worker_id not in self.workers:
                worker_record = current_workerlist[worker_id]
                self.workers[worker_id] = self._create_worker_proxy(
                    worker_record)
            # Otherwise it hasn't changed.

    def _remove_worker_proxy(self, worker_id):
        worker = self.workers.get(worker_id)
        if not worker:
            return
        forwarder = worker["forwarder"]
        forwarder.stop()

    def _create_worker_proxy(self, worker_record):
        host = worker_record["host"]
        ipaddr = ipaddress.IPAddress(host)
        port = worker_record["services"].get("bokeh")
        if not port:
            return None
        if ipaddr.is_local():
            proxy = None
            local_port = port
        else:
            proxy = Forwarder(host, port, ioloop=self.ioloop)
            proxy.start()
            local_port = proxy.get_port()
        url = get_proxy_url(local_port)
        worker = {"forwarder": proxy,
                  "url": url,
                  "local_port": local_port}
        return worker

    def get_proxies(self, workers):
        """Returns a dict of worker endpoints as keys, mapped to a dict
        containing the worker proxy url and local port it's mapped to.
        """
        rval = {}
        if not workers:
            workers = list(self.workers.keys())
        if workers:
            if type(workers) is str:
                workers = [workers]
        for worker in workers:
            rval[worker] = {"url": self.workers[worker]["url"],
                            "local_port": self.workers[worker]["local_port"]}
        return rval

    def __repr__(self):
        s = "ClusterProxy {name}:".format(name=get_hostname())
        s += "\n  Scheduler: {url}".format(url=self.scheduler_url)
        sw = self.workers
        if sw:
            s = s+"\n  Workers:"
        for worker in sw:
            s += "\n    {worker}: {url}".format(worker=worker,
                                                url=sw[worker]["url"])
        return s

    def _repr_html_(self):
        s = "<h3>ClusterProxy {name}</h3>".format(name=get_hostname())
        s += "<h4>Scheduler: {url}</h4>".format(url=self.scheduler_url)
        if len(self.workers) > 0:
            s += "<h4>Workers</h4><dl>"
            sw = self.workers
            for worker in sw:
                s += "<dt>{w}</dt><dd>{u}</dd>".format(w=worker,
                                                       u=sw[worker]["url"])
            s += "</dl>"
        return s



from .utils import get_proxy_url
import dask
from dask.distributed import Client, sync
import logging
import os
from tornado import gen
from tornado.ioloop import IOLoop

logger = logging.getLogger(__name__)


class LSSTDaskClient(Client):
    """This uses the proxy info to provide an externally-reachable dashboard
    URL for the Client.  Other than that it does nothing differently from the
    standard dask.distributed.Client.  It assumes the LSST JupyterLab
    environment.
    """
    proxy_url = None

    @gen.coroutine
    def _update_scheduler_info(self):
        if self.status not in ('running', 'connecting'):
            return
        try:
            self._scheduler_identity = yield self.scheduler.identity()
        except EnvironmentError:
            logger.debug("Not able to query scheduler for identity")
        srv = self._scheduler_identity.get("services")
        if srv:
            self.proxy_url = get_proxy_url(srv.get("bokeh"))

    def __repr__(self):
        # Note: avoid doing I/O here...
        info = self._scheduler_identity
        addr = self.proxy_url or info.get('address')
        if addr:
            workers = info.get('workers', {})
            nworkers = len(workers)
            ncores = sum(w['ncores'] for w in workers.values())
            return '<%s: scheduler=%r processes=%d cores=%d>' % (
                self.__class__.__name__, addr, nworkers, ncores)
        elif self.scheduler is not None:
            return '<%s: scheduler=%r>' % (
                self.__class__.__name__, self.scheduler.address)
        else:
            return '<%s: not connected>' % (self.__class__.__name__,)

    def _repr_html_(self):
        if (self.cluster and hasattr(self.cluster, 'scheduler') and
                self.cluster.scheduler):
            info = self.cluster.scheduler.identity()
            scheduler = self.cluster.scheduler
        elif (self._loop_runner.is_started() and
                self.scheduler and
                not (self.asynchronous and self.loop is IOLoop.current())):
            info = sync(self.loop, self.scheduler.identity)
            scheduler = self.scheduler
        else:
            info = False
            scheduler = self.scheduler

        if scheduler is not None:
            text = ("<h3>Client</h3>\n"
                    "<ul>\n"
                    "  <li><b>Scheduler: </b>%s\n") % scheduler.address
        else:
            text = ("<h3>Client</h3>\n"
                    "<ul>\n"
                    "  <li><b>Scheduler: not connected</b>\n")
        sched_addr = self.proxy_url or scheduler.address
        if info and 'bokeh' in info['services']:
            protocol, rest = sched_addr.split('://')
            port = info['services']['bokeh']
            if protocol == 'inproc':
                host = 'localhost'
            else:
                host = rest.split(':')[0]
            template = dask.config.get('distributed.dashboard.link')
            address = template.format(host=host, port=port, **os.environ)
            text += "  <li><b>Dashboard: </b>"
            text += "<a href='%(web)s' target='_blank'>%(web)s</a>\n" % {
                'web': address}

        text += "</ul>\n"

        if info:
            workers = len(info['workers'])
            cores = sum(w['ncores'] for w in info['workers'].values())
            memory = sum(w['memory_limit'] for w in info['workers'].values())
            # memory = format_bytes(memory)
            text2 = ("<h3>Cluster</h3>\n"
                     "<ul>\n"
                     "  <li><b>Workers: </b>%d</li>\n"
                     "  <li><b>Cores: </b>%d</li>\n"
                     "  <li><b>Memory: </b>%s</li>\n"
                     "</ul>\n") % (workers, cores, memory)

            return ('<table style="border: 2px solid white;">\n'
                    '<tr>\n'
                    '<td style="vertical-align: top; '
                    'border: 0px solid white">\n%s</td>\n'
                    '<td style="vertical-align: top; '
                    'border: 0px solid white">\n%s</td>\n'
                    '</tr>\n</table>') % (text, text2)

        else:
            return text
